import collections

from keg_storage import cli


class Storage:
    """A proxy and management object for storage backends."""

    _interaces = None

    def __init__(self, app=None, cli_group_name='storage'):
        self.cli_group_name = cli_group_name

        if app:
            self.init_app(app)

    def init_app(self, app):
        app.storage = self

        self._interfaces = collections.OrderedDict(
            (params['name'], interface(**params))
            for interface, params in app.config['STORAGE_PROFILES']
        )
        self.interface = next(iter(self._interfaces)) if self._interfaces else None

        self.cli_group = self.init_cli(app)

    def get_interface(self, interface=None):
        return self._interfaces[interface or self.interface]

    def init_cli(self, app):
        cli.add_cli_to_app(app, self.cli_group_name)



import datetime
import io

import arrow
from botocore.exceptions import ClientError
import pytest
from mock import mock

import keg_storage.backends as backends
from keg_storage.backends.base import (
    FileNotFoundInStorageError,
    ListEntry,
    FileMode,
)


@mock.patch('keg_storage.backends.s3.boto3', autospec=True, spec_set=True)
class TestS3Storage:
    def test_init_sets_up_correctly(self, m_boto):
        s3 = backends.S3Storage('bucket', aws_region='us-east-1', aws_access_key_id='key',
                                aws_secret_access_key='secret', name='test')
        assert s3.name == 'test'
        assert s3.bucket == 'bucket'

        m_boto.session.Session.assert_called_once_with(
            aws_access_key_id='key',
            aws_secret_access_key='secret',
            profile_name=None,
            region_name='us-east-1'
        )

    def test_list(self, m_boto):
        s3 = backends.S3Storage('bucket', aws_region='us-east-1')
        s3.client.list_objects_v2.return_value = {
            'IsTruncated': False,
            'Contents': [
                {
                    'Key': 'file-1.wps',
                    'LastModified': datetime.datetime(2019, 8, 26, 15, 30, 1),
                    'Size': 10 * 1024
                },
                {
                    'Key': 'file-2.rm',
                    'LastModified': datetime.datetime(2019, 8, 26, 15, 30, 2),
                    'Size': 20 * 1024
                },
            ]
        }

        results = s3.list('foo/bar')

        s3.client.list_objects_v2.assert_called_once_with(
            Bucket='bucket',
            Prefix='foo/bar'
        )

        assert results == [
            ListEntry(
                name='file-1.wps',
                last_modified=arrow.get(2019, 8, 26, 15, 30, 1),
                size=10240
            ),
            ListEntry(
                name='file-2.rm',
                last_modified=arrow.get(2019, 8, 26, 15, 30, 2),
                size=20480
            )
        ]

    def test_list_pagenated(self, m_boto):
        s3 = backends.S3Storage('bucket', aws_region='us-east-1')

        s3.client.list_objects_v2.side_effect = [
            {
                'IsTruncated': True,
                'NextContinuationToken': 'next-token-1',
                'Contents': [
                    {
                        'Key': 'file-1.wps',
                        'LastModified': datetime.datetime(2019, 8, 26, 15, 30, 1),
                        'Size': 10 * 1024
                    },
                    {
                        'Key': 'file-2.rm',
                        'LastModified': datetime.datetime(2019, 8, 26, 15, 30, 2),
                        'Size': 20 * 1024
                    },
                ]
            },
            {
                'IsTruncated': True,
                'NextContinuationToken': 'next-token-2',
                'Contents': [
                    {
                        'Key': 'file-3.rar',
                        'LastModified': datetime.datetime(2019, 8, 26, 15, 30, 3),
                        'Size': 5 * 1024
                    },
                ]
            },
            {
                'IsTruncated': False,
                'Contents': [
                    {
                        'Key': 'file-4.rtf',
                        'LastModified': datetime.datetime(2019, 8, 26, 15, 30, 4),
                        'Size': 1024
                    },
                ]
            },

        ]

        results = s3.list('foo/bar')

        assert s3.client.list_objects_v2.call_args_list == [
            mock.call(Bucket='bucket', Prefix='foo/bar'),
            mock.call(Bucket='bucket', Prefix='foo/bar', ContinuationToken='next-token-1'),
            mock.call(Bucket='bucket', Prefix='foo/bar', ContinuationToken='next-token-2'),
        ]

        assert results == [
            ListEntry(
                name='file-1.wps',
                last_modified=arrow.get(2019, 8, 26, 15, 30, 1),
                size=10240
            ),
            ListEntry(
                name='file-2.rm',
                last_modified=arrow.get(2019, 8, 26, 15, 30, 2),
                size=20480
            ),
            ListEntry(
                name='file-3.rar',
                last_modified=arrow.get(2019, 8, 26, 15, 30, 3),
                size=5120
            ),
            ListEntry(
                name='file-4.rtf',
                last_modified=arrow.get(2019, 8, 26, 15, 30, 4),
                size=1024
            ),
        ]

    def test_delete(self, m_boto):
        s3 = backends.S3Storage('bucket', aws_region='us-east-1')

        s3.delete('foo/bar')

        s3.client.delete_object.assert_called_once_with(
            Bucket='bucket',
            Key='foo/bar'
        )

    def test_open_read(self, m_boto):
        s3 = backends.S3Storage('bucket', aws_region='us-east-1')

        result = s3.open('foo/bar', FileMode.read)
        assert isinstance(result, backends.s3.S3Reader)
        s3.client.get_object.assert_called_once_with(Bucket='bucket', Key='foo/bar')

    def test_open_write(self, m_boto):
        s3 = backends.S3Storage('bucket', aws_region='us-east-1')

        result = s3.open('foo/bar', FileMode.write)
        assert isinstance(result, backends.s3.S3Writer)

    def test_open_read_write(self, m_boto):
        s3 = backends.S3Storage('bucket', aws_region='us-east-1')
        with pytest.raises(NotImplementedError) as exc:
            s3.open('foo/bar', FileMode.read | FileMode.write)
        assert str(exc.value) == 'Read+write mode not supported by the S3 backend'

        with pytest.raises(ValueError) as exc:
            s3.open('foo/bar', FileMode(0))
        assert (
            str(exc.value) == 'Unsupported mode. Accepted modes are FileMode.read or FileMode.write'
        )

    def test_read_operations(self, m_boto):
        s3 = backends.S3Storage('bucket', aws_region='us-east-1')
        body_obj = io.BytesIO(b'a' * 100)
        s3.client.get_object.return_value = {
            'Body': body_obj
        }

        with s3.open('foo/bar', FileMode.read) as fp:
            assert fp.read(1) == b'a'
            assert fp.read(2) == b'aa'
            assert fp.read(37) == b'a' * 37
            assert fp.read(50) == b'a' * 50

        s3.client.get_object.assert_called_once_with(Bucket='bucket', Key='foo/bar')
        assert fp.reader.closed is True

    def test_read_not_found(self, m_boto):
        s3 = backends.S3Storage('bucket', aws_region='us-east-1')
        s3.client.get_object.side_effect = ClientError({'Error': {'Code': 'NoSuchKey'}}, 'foo')

        with pytest.raises(FileNotFoundInStorageError) as exc:
            s3.open('foo/bar', FileMode.read)

        assert str(exc.value.filename) == 'foo/bar'
        assert str(exc.value.storage_type) == 'S3Storage'

    def test_write_operations(self, m_boto):
        m_client = mock.MagicMock()
        m_client.create_multipart_upload.return_value = {'UploadId': 'upload-id'}
        m_client.upload_part.side_effect = [{'ETag': f'etag-{x}'} for x in range(5)]

        with backends.s3.S3Writer('bucket', 'foo/bar', m_client, chunk_size=100) as fp:
            m_client.create_multipart_upload.assert_not_called()

            fp.write(b'a')
            m_client.create_multipart_upload.assert_not_called()
            m_client.upload_part.assert_not_called()

            fp.write(b'b' * 100)
            m_client.create_multipart_upload.assert_called_once_with(
                Bucket='bucket',
                Key='foo/bar'
            )
            m_client.upload_part.assert_called_once_with(
                Bucket='bucket',
                Key='foo/bar',
                PartNumber=1,
                UploadId='upload-id',
                Body=b'a' + b'b' * 99
            )

            # test a write bigger than the buffer size
            fp.write(b'c' * 200)
            m_client.upload_part.assert_any_call(
                Bucket='bucket',
                Key='foo/bar',
                PartNumber=2,
                UploadId='upload-id',
                Body=b'b' + b'c' * 99
            )
            m_client.upload_part.assert_any_call(
                Bucket='bucket',
                Key='foo/bar',
                PartNumber=3,
                UploadId='upload-id',
                Body=b'c' * 100
            )

            m_client.complete_multipart_upload.assert_not_called()

        m_client.upload_part.assert_called_with(
            Bucket='bucket',
            Key='foo/bar',
            PartNumber=4,
            UploadId='upload-id',
            Body=b'c'
        )

        m_client.complete_multipart_upload.assert_called_once_with(
            Bucket='bucket',
            Key='foo/bar',
            UploadId='upload-id',
            MultipartUpload={
                'Parts': [
                    {
                        'PartNumber': 1,
                        'ETag': 'etag-0',
                    },
                    {
                        'PartNumber': 2,
                        'ETag': 'etag-1',
                    },
                    {
                        'PartNumber': 3,
                        'ETag': 'etag-2',
                    },
                    {
                        'PartNumber': 4,
                        'ETag': 'etag-3',
                    },
                ]
            }
        )

    def test_write_abort(self, m_boto):
        m_client = mock.MagicMock()
        m_client.create_multipart_upload.return_value = {'UploadId': 'upload-id'}
        m_client.upload_part.return_value = {'ETag': 'etag-0'}

        with backends.s3.S3Writer('bucket', 'foo/bar', m_client, chunk_size=100) as fp:
            fp.write(b'a' * 99)  # fill the buffer but don't trigger a flush
            fp.abort()
        # We haven't created the upload yet so aborting should do nothing
        m_client.abort_multipart_upload.assert_not_called()

        with backends.s3.S3Writer('bucket', 'foo/bar', m_client, chunk_size=100) as fp:
            fp.write(b'a' * 100)  # Force a buffer flush
            fp.abort()

        m_client.complete_multipart_upload.assert_not_called()
        m_client.abort_multipart_upload.assert_called_once_with(
            Bucket='bucket',
            Key='foo/bar',
            UploadId='upload-id'
        )

    def test_write_flushes(self, m_boto):
        m_client = mock.MagicMock()
        m_client.create_multipart_upload.return_value = {'UploadId': 'upload-id'}
        m_client.upload_part.return_value = {'ETag': 'etag-0'}

        with backends.s3.S3Writer('bucket', 'foo/bar', m_client, chunk_size=100) as fp:
            fp.write(b'a' * 99)
        m_client.create_multipart_upload.assert_called()
        m_client.upload_part.assert_called()
        m_client.complete_multipart_upload.assert_called()



from __future__ import absolute_import

from keg.app import Keg

from keg_storage_ta.extensions import storage


class KegStorageTestApp(Keg):
    import_name = 'keg_storage_ta'
    db_enabled = False
    keyring_enable = False

    def on_init_complete(self):
        storage.init_app(self)
        return self


if __name__ == '__main__':
    from keg_storage_ta import app
    app.KegStorageTestApp.cli.main()



import click
import functools
from flask.cli import with_appcontext
import humanize

from keg_storage.backends.base import FileNotFoundInStorageError
from keg_storage import utils


@click.group('_storage')
@click.option('--location')
@with_appcontext
@click.pass_context
def storage(ctx, location):
    from flask import current_app
    location = location or current_app.config.get('KEG_STORAGE_DEFAULT_LOCATION')
    if not location:
        click.echo('No location given and no default was configured.')
        ctx.abort()
    try:
        ctx.obj.data['storage'] = current_app.storage.get_interface(location)
    except KeyError:
        click.echo('The location {} does not exist. '
                   'Pass --location or change your configuration.'.format(location))
        ctx.abort()


def handle_not_found(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            func(*args, **kwargs)
        except FileNotFoundInStorageError as e:
            raise click.FileError(
                e.filename,
                hint="Not found in {}.".format(str(e.storage_type))
            )

    return wrapper


@storage.command('list')
@click.option('--simple', is_flag=True)
@click.argument('path', default='/')
@click.pass_context
def storage_list(ctx, path, simple):
    objs = ctx.obj.data['storage'].list(path)

    def fmt(item):
        fmt_str = '{name}' if simple else '{date}\t{size}\t{name}'
        keys = {
            'date': humanize.naturaldate(item.last_modified),
            'name': item.name,
            'size': humanize.naturalsize(item.size, gnu=True),
        }
        return fmt_str.format(**keys)

    lines = [fmt(item) for item in objs]
    click.echo("\n".join(lines))


@storage.command('get')
@click.argument('path')
@click.argument('dest', default='')
@click.pass_context
@handle_not_found
def storage_get(ctx, path, dest):
    if dest == '':
        dest = path.split('/')[-1]

    ctx.obj.data['storage'].get(path, dest)
    click.echo("Downloaded {path} to {dest}.".format(path=path, dest=dest))


@storage.command('put')
@click.argument('path')
@click.argument('key')
@click.pass_context
def storage_put(ctx, path, key):
    ctx.obj.data['storage'].put(path, key)
    click.echo("Uploaded {path} to {key}.".format(key=key, path=path))


@storage.command('delete')
@click.argument('path')
@click.pass_context
@handle_not_found
def storage_delete(ctx, path):
    ctx.obj.data['storage'].delete(path)
    click.echo("Deleted {path}.".format(path=path))


@storage.command('link_for')
@click.argument('path')
@click.option('--expiration', '-e', default=3600,
              help="Expiration time (in seconds) of the link, defaults to 1 hours")
@click.pass_context
def storage_link_for(ctx, path, expiration):

    try:
        retval = ctx.obj.data['storage'].link_for(path, expiration)
    except Exception as e:
        click.echo(str(e))
        ctx.abort()

    click.echo("{data}".format(data=retval))


@storage.command('reencrypt')
@click.argument('path')
@click.pass_context
@handle_not_found
def storage_reencrypt(ctx, path):
    old_key = click.prompt('Old Key', hide_input=True).encode('ascii')
    new_key = click.prompt('New Key', hide_input=True).encode('ascii')

    utils.reencrypt(ctx.obj.data['storage'], path, old_key, new_key)
    click.echo('Re-encrypted {path}'.format(path=path))


def add_cli_to_app(app, cli_group_name):
    app.cli.add_command(storage, name=cli_group_name)



import keg_storage


class DefaultProfile(object):
    # This just gets rid of warnings on the console.
    KEG_KEYRING_ENABLE = False

    SITE_NAME = 'Keg Storage Demo'
    SITE_ABBR = 'KS Demo'

    STORAGE_PROFILES = [
        (keg_storage.S3Storage, {
            'name': 'storage.s3',
            'bucket': 'storage.test',
            'aws_region': 'us-east-1',
            'aws_access_key_id': 'access-key-id',
            'aws_secret_access_key': 'secret-key',
        }),
        (keg_storage.SFTPStorage, {
            'name': 'storage.sftp',
            'host': 'example.com',
            'username': 'john.doe',
            'key_filename': '/key/path',
            'known_hosts_fpath': '/known/hosts/path',
            'allow_agent': False,
            'look_for_keys': False,
        })
    ]
    # KEG_STORAGE_DEFAULT_LOCATION = 'storage.s3'


class TestProfile(object):
    pass



import os

import arrow
import click
import pytest

import keg_storage
from keg_storage.backends.base import (
    ListEntry,
    FileMode,
    RemoteFile,
)
from keg_storage.cli import handle_not_found


class FakeRemoteFile(RemoteFile):
    iter_chunk_size = 20

    def __init__(self, path, mode):
        super().__init__(mode)
        self.file = open(path, str(self.mode))

    def read(self, size):
        return self.file.read(size)

    def write(self, data):
        return self.file.write(data)

    def close(self):
        self.file.close()


class FakeBackend(keg_storage.StorageBackend):
    def __init__(self, dir):
        super().__init__()
        self.base_dir = dir

    def list(self, path):
        results = []
        path = os.path.join(self.base_dir, path)
        for dirent in os.scandir(path):
            stat = dirent.stat()
            results.append(
                ListEntry(
                    name=dirent.name,
                    last_modified=arrow.get(stat.st_mtime),
                    size=stat.st_size
                )
            )
        return results

    def open(self, path, mode):
        path = os.path.join(self.base_dir, path)
        return FakeRemoteFile(path, mode)

    def delete(self, path):
        path = os.path.join(self.base_dir, path)
        os.unlink(path)


class TestStorageBackend:

    def test_methods_not_implemented(self):

        interface = keg_storage.StorageBackend()

        cases = {
            interface.list: ('path',),
            interface.delete: ('path',),
            interface.open: ('path', FileMode.read),
        }

        for method, args in cases.items():
            with pytest.raises(NotImplementedError):
                method(*args)

    def test_get(self, tmp_path):
        remote = tmp_path / 'remote'
        local = tmp_path / 'local'

        remote.mkdir()
        local.mkdir()

        data = b'a' * 15000000
        with (remote / 'input_file.txt').open('wb') as fp:
            fp.write(data)

        interface = FakeBackend(str(remote))

        output_path = local / 'output_file.txt'
        interface.get('input_file.txt', str(output_path))

        with output_path.open('rb') as of:
            assert of.read() == data

    def test_put(self, tmp_path):
        remote = tmp_path / 'remote'
        local = tmp_path / 'local'

        remote.mkdir()
        local.mkdir()

        data = b'a' * 15000000
        input_path = local / 'input_file.txt'
        with input_path.open('wb') as fp:
            fp.write(data)

        interface = FakeBackend(str(remote))
        interface.put(str(input_path), 'output_file.txt')

        with (remote / 'output_file.txt').open('rb') as of:
            assert of.read() == data

    def test_str(self, tmp_path):
        interface = FakeBackend(str(tmp_path))
        assert str(interface) == 'FakeBackend'

    def test_remote_file_iter_chunks(self, tmp_path):
        file_path = tmp_path / 'test_file.txt'
        with file_path.open('wb') as fp:
            fp.write(
                b'a' * 100 +
                b'b' * 100 +
                b'c' * 5
            )

        file = FakeRemoteFile(str(file_path), FileMode.read)
        chunks = list(file.iter_chunks(100))
        assert chunks == [
            b'a' * 100,
            b'b' * 100,
            b'c' * 5
        ]

    def test_remote_file_closes_on_delete(self, tmp_path):
        file_path = tmp_path / 'test_file.txt'
        file = FakeRemoteFile(str(file_path), FileMode.write)

        real_file = file.file
        assert real_file.closed is False

        del file
        assert real_file.closed is True

    def test_remote_file_context_manager(self, tmp_path):
        file_path = tmp_path / 'test_file.txt'
        file = FakeRemoteFile(str(file_path), FileMode.write)
        real_file = file.file

        with file:
            assert real_file.closed is False
        assert real_file.closed is True

    def test_remote_file_iter(self, tmp_path):
        file_path = tmp_path / 'test_file.txt'
        with file_path.open('wb') as fp:
            fp.write(
                b'a' * 20 +
                b'b' * 20 +
                b'c' * 20
            )

        file = FakeRemoteFile(str(file_path), FileMode.read)
        chunks = list(file)
        assert chunks == [
            b'a' * 20,
            b'b' * 20,
            b'c' * 20
        ]


class TestFileNotFoundException:
    def test_click_wrapper(self, tmp_path):
        backend = FakeBackend(str(tmp_path))

        @handle_not_found
        def test_func():
            raise keg_storage.FileNotFoundInStorageError(backend, 'foo')

        with pytest.raises(click.FileError) as exc_info:
            test_func()
        assert exc_info.value.filename == 'foo'
        assert exc_info.value.message == 'Not found in FakeBackend.'


class TestFileMode:
    def test_str(self):
        assert str(FileMode.read) == 'rb'
        assert str(FileMode.write) == 'wb'
        assert str(FileMode.read | FileMode.write) == 'rwb'

    def test_as_mode(self):
        assert FileMode.as_mode(FileMode.read) == FileMode.read

        assert FileMode.as_mode('r') == FileMode.read
        assert FileMode.as_mode('w') == FileMode.write
        assert FileMode.as_mode('rw') == FileMode.read | FileMode.write
        assert FileMode.as_mode('wr') == FileMode.read | FileMode.write
        assert FileMode.as_mode('rb') == FileMode.read

        with pytest.raises(ValueError) as exc:
            FileMode.as_mode(1)
        assert str(exc.value) == 'as_mode() accepts only FileMode or str arguments'



import contextlib
import logging
import typing

import arrow
from paramiko import SSHClient

from .base import (
    StorageBackend,
    ListEntry,
    FileMode,
    RemoteFile,
)

log = logging.getLogger(__name__)


class SFTPRemoteFile(RemoteFile):
    def __init__(self, mode, path, client):
        super().__init__(mode)
        self.path = path
        self.client = client
        self.file = None

        self.sftp = client.open_sftp()
        self.file = self.sftp.open(path, str(mode))

    def read(self, size: int):
        if not (self.mode & FileMode.read):
            raise IOError('File not opened for reading')
        return self.file.read(size)

    def close(self):
        # File may actually be none since the open operation in the constructor may have failed
        # before assigning the value
        if self.file is not None:
            self.file.close()
        self.client.close()

    def write(self, data: bytes):
        if not (self.mode & FileMode.write):
            raise IOError('File not opened for writing')
        return self.file.write(data)


class SFTPStorage(StorageBackend):
    def __init__(
            self,
            host,
            username,
            key_filename,
            known_hosts_fpath,
            allow_agent=False,
            look_for_keys=False,
            name='sftp'
    ):
        super().__init__()
        self.host = host
        self.username = username
        self.key_filename = key_filename
        self.known_hosts_fpath = known_hosts_fpath
        self.allow_agent = allow_agent
        self.look_for_keys = look_for_keys
        self.name = name

    def create_client(self):
        client = SSHClient()
        client.load_system_host_keys(self.known_hosts_fpath)

        client.connect(
            self.host,
            username=self.username,
            key_filename=self.key_filename,
            allow_agent=self.allow_agent,
            look_for_keys=self.look_for_keys,
        )

        return client

    @contextlib.contextmanager
    def connection(self):
        with self.create_client() as client:
            sftp = client.open_sftp()
            yield sftp

    def list(self, path: str):
        with self.connection() as conn:
            return [
                ListEntry(
                    name=x.filename,
                    last_modified=arrow.get(x.st_mtime),
                    size=x.st_size
                )
                for x in conn.listdir_attr(path)
            ]

    def open(self, path: str, mode: typing.Union[FileMode, str]):
        mode = FileMode.as_mode(mode)

        # SFTPRemoteFile is responsible for closing the client connection
        client = self.create_client()
        return SFTPRemoteFile(mode, path, client)

    def delete(self, path: str):
        log.info("Deleting remote file '%s'", path)

        with self.connection() as conn:
            conn.remove(path)



import io

import pytest
from mock import mock

try:
    import keg_elements.crypto as ke_crypto
except ImportError:
    ke_crypto = None

from keg_storage import utils

DEFAULT_PLAINTEXT = b'data' * 1024 * 1024


@pytest.mark.skipif(ke_crypto is None, reason='Only run with keg elements installed')
class TestReencrypt:
    def mock_storage(self, old_ciphertext, new_ciphertext):
        m_storage = mock.MagicMock()

        def fake_get(src, dest):
            with open(dest, 'wb') as fp:
                fp.write(old_ciphertext)

        m_storage.get.side_effect = fake_get

        def fake_put(src, dest):
            with open(src, 'rb') as fp:
                new_ciphertext.write(fp.read())

        m_storage.put.side_effect = fake_put
        return m_storage

    def create_mock_storage(self, old, new, plaintext=DEFAULT_PLAINTEXT):
        n_ct = io.BytesIO()
        o_ct = b''.join(ke_crypto.encrypt_fileobj(old, io.BytesIO(plaintext)))
        return self.mock_storage(o_ct, n_ct), o_ct, n_ct

    def test_reencrypt_with_bad_keys_throws_key_error(self):
        with pytest.raises(utils.EncryptionKeyException):
            utils.reencrypt(None, 'file', b'', b'')

        with pytest.raises(utils.EncryptionKeyException):
            utils.reencrypt(None, 'file', b'a' * 32, b'')

        with pytest.raises(utils.EncryptionKeyException):
            utils.reencrypt(None, 'file', [b'a', b'b' * 20] * 32, b'')

    def test_reencrypt_takes_multiple_key(self):
        k1, k2, k3, = b'a' * 32, b'b' * 32, b'c' * 32

        store, o_ct, n_ct = self.create_mock_storage(k1, k3)
        utils.reencrypt(store, 'file', [k1, k2], k3)

        n_ct.seek(0)
        new_plaintext = io.BytesIO()
        ke_crypto.decrypt_fileobj(k3, n_ct, new_plaintext, chunksize=1024)

        assert new_plaintext.getvalue() == DEFAULT_PLAINTEXT

    def test_reencrypt_with_single_key(self):
        k1, k2 = b'a' * 32, b'b' * 32

        store, o_ct, n_ct = self.create_mock_storage(k1, k2)
        utils.reencrypt(store, 'file', k1, k2)

        assert store.get.call_count == 1
        args, kwargs = store.get.call_args
        assert len(args) == 2
        assert args[0] == 'file'

        assert store.put.call_count == 1
        args, kwargs = store.put.call_args
        assert len(args) == 2
        assert args[1] == 'file'

        n_ct.seek(0)
        new_plaintext = io.BytesIO()
        ke_crypto.decrypt_fileobj(k2, n_ct, new_plaintext, chunksize=1024)

        assert new_plaintext.getvalue() == DEFAULT_PLAINTEXT

    def test_with_bad_key(self):
        k1, k2, k3, = b'a' * 32, b'b' * 32, b'c' * 32
        store, o_ct, n_ct = self.create_mock_storage(k2, k3)

        with pytest.raises(utils.DecryptionException):
            utils.reencrypt(store, 'file', k1, k3)


@pytest.mark.skipif(ke_crypto is not None, reason='Only run with keg elements not installed')
def test_reencrypt():
    with pytest.raises(utils.MissingDependencyException) as exc:
        utils.reencrypt(None, '', '', '')
    assert str(exc.value) == 'Keg Elements is required for crypto operations'



import arrow
import pytest
from blazeutils.containers import LazyDict
from unittest import mock
import wrapt

import keg_storage
from keg_storage.backends.base import (
    ListEntry,
    FileMode,
)
from keg_storage.backends.sftp import SFTPRemoteFile


def sftp_mocked(**kwargs):
    @wrapt.decorator(adapter=lambda self: None)
    def wrapper(wrapped, instance, args, _kwargs):

        @mock.patch('keg_storage.sftp.log', autospec=True, spec_set=True)
        def run_test(m_log):
            m_client = mock.MagicMock(
                spec=keg_storage.sftp.SSHClient,
                spec_set=keg_storage.sftp.SSHClient
            )

            class FakeSFTPStorage(keg_storage.sftp.SFTPStorage):
                def create_client(self):
                    return m_client

            fake_sftp = FakeSFTPStorage(
                host=kwargs.get('host', 'foo'),
                username=kwargs.get('username', 'bar'),
                key_filename=kwargs.get('key_filename'),
                known_hosts_fpath=kwargs.get('known_hosts_fpath', 'known_hosts'),
            )

            m_sftp = mock.MagicMock()

            m_client.__enter__.return_value = m_client
            m_client.open_sftp.return_value = m_sftp
            wrapped(
                sftp=fake_sftp,
                m_sftp=m_sftp,
                m_log=m_log
            )

        return run_test()
    return wrapper


class TestSFTPStorage:

    @sftp_mocked()
    def test_sftp_list_files(self, sftp, m_sftp, m_log):
        files = [
            LazyDict(filename='a.txt', st_mtime=1564771623, st_size=128),
            LazyDict(filename='b.pdf', st_mtime=1564771638, st_size=32768),
            LazyDict(filename='more.txt', st_mtime=1564771647, st_size=100)
        ]
        m_sftp.listdir_attr.return_value = files
        assert sftp.list('.') == [
            ListEntry(name='a.txt', last_modified=arrow.get(1564771623), size=128),
            ListEntry(name='b.pdf', last_modified=arrow.get(1564771638), size=32768),
            ListEntry(name='more.txt', last_modified=arrow.get(1564771647), size=100),
        ]
        assert m_log.info.mock_calls == []

    @sftp_mocked()
    def test_sftp_delete_file(self, sftp, m_sftp, m_log):
        sftp.delete('/tmp/abc/baz.txt')
        m_sftp.remove.assert_called_once_with('/tmp/abc/baz.txt')
        m_log.info.assert_called_once_with("Deleting remote file '%s'", '/tmp/abc/baz.txt')

    @sftp_mocked()
    def test_open(self, sftp, m_sftp, m_log):
        file = sftp.open('/tmp/foo.txt', FileMode.read)
        assert isinstance(file, SFTPRemoteFile)

        assert file.mode == FileMode.read
        assert file.path == '/tmp/foo.txt'
        assert file.sftp is m_sftp

        m_sftp.open.assert_called_once_with('/tmp/foo.txt', 'rb')

    @sftp_mocked()
    def test_read_operations(self, sftp, m_sftp, m_log):
        m_file = m_sftp.open.return_value
        m_file.read.return_value = b'some data'

        with sftp.open('/tmp/foo.txt', FileMode.read) as file:
            assert file.read(4) == b'some data'
            m_file.read.assert_called_once_with(4)
            m_file.close.assert_not_called()
        m_file.close.assert_called_once_with()

    @sftp_mocked()
    def test_read_not_permitted(self, sftp, m_sftp, m_log):
        with sftp.open('/tmp/foo.txt', FileMode.write) as file:
            with pytest.raises(IOError) as exc:
                file.read(1)
        assert str(exc.value) == 'File not opened for reading'

    @sftp_mocked()
    def test_write_operations(self, sftp, m_sftp, m_log):
        m_file = m_sftp.open.return_value
        with sftp.open('/tmp/foo.txt', FileMode.write) as file:
            file.write(b'some data')
            m_file.write.assert_called_once_with(b'some data')
            m_file.close.assert_not_called()
        m_file.close.assert_called_once_with()

    @sftp_mocked()
    def test_write_not_permitted(self, sftp, m_sftp, m_log):
        with sftp.open('/tmp/foo.txt', FileMode.read) as file:
            with pytest.raises(IOError) as exc:
                file.write(b'')
        assert str(exc.value) == 'File not opened for writing'



import typing

import arrow
import boto3
from botocore.exceptions import ClientError

from .base import (
    StorageBackend,
    FileNotFoundInStorageError,
    ListEntry,
    RemoteFile,
    FileMode,
)


class S3FileBase(RemoteFile):
    """
    Read and write operations for S3 are very different so individual subclasses are used for each.
    Read+Write mode is not available for this backend.
    """
    def __init__(self, mode, bucket, filename, client):
        self.s3 = client
        self.bucket = bucket
        self.filename = filename
        super().__init__(mode)


class S3Reader(S3FileBase):
    def __init__(self, bucket, filename, client):
        super().__init__(FileMode.read, bucket, filename, client)
        self.reader = None

        obj = self.s3.get_object(Bucket=self.bucket, Key=self.filename)
        self.reader = obj['Body']

    def read(self, size: int):
        return self.reader.read(size)

    def close(self):
        # Reader may be None if an exception is thrown in the constructor
        if self.reader is not None:
            self.reader.close()


class S3Writer(S3FileBase):
    """
    Writes to S3 are quite a bit more complicated than reads. To support large files, we cannot
    write in a single operation and the API does not encourage streaming writes so we make use of
    the multipart API methods.

    The process can be summarized as:
        * Create a multipart upload and get an upload key to use with subsequent calls.
        * Upload "parts" of the file using the upload key and get back an ID for each part.
        * Combine the parts using the upload key and all the part IDs from the above steps.

    The chunked nature of the uploads should be mostly invisible to the caller since S3Writer
    maintains a local buffer.

    Because creating a multipart upload itself has an actual cost and there is no guarantee that
    anything will actually be written, we initialize the multipart upload lazily.
    """
    def __init__(self, bucket, filename, client, chunk_size=10 * 1024 * 1024):
        super().__init__(FileMode.write, bucket, filename, client)
        # The upload key that we will get when we intialize the multipart upload
        self.multipart_id = None

        # The IDs for each part we upload in the order they should be combined
        self.part_ids = []

        # A local buffer to limit the number of parts we need to create
        self.buffer = bytearray()

        # The maximum size of an uploaded part
        self.chunk_size = chunk_size

    def _flush_buffer(self):
        """
        Upload the contents of the local buffer to an S3 part.
        """
        if not self.buffer:
            # If the buffer is already empty, do nothing
            return

        if self.multipart_id is None:
            # Create the multipart upload if it has not been initialized yet
            self._init_multipart()

        # Upload the first `chunk_size` bytes from the buffer to create a part
        body = self.buffer[:self.chunk_size]
        part = self.s3.upload_part(
            Bucket=self.bucket,
            Key=self.filename,
            PartNumber=len(self.part_ids) + 1,
            UploadId=self.multipart_id,
            Body=body
        )

        # Store the resulting part ID to recombine later
        self.part_ids.append(part['ETag'])

        # Cycle the buffer
        self.buffer = self.buffer[self.chunk_size:]

    def _init_multipart(self):
        """
        Create the S3 multipart upload
        """
        mpu = self.s3.create_multipart_upload(Bucket=self.bucket, Key=self.filename)
        self.multipart_id = mpu['UploadId']

    def _finalize_multipart(self):
        """
        Recombine all the uploaded parts into a single S3 object.
        """
        # If no data was uploaded, we don't need to do anything else
        if not self.part_ids:
            return

        # Specify how to combine the parts. The PartNumber indicates the order and must be >= 1
        part_info = {
            'Parts': [
                {
                    'PartNumber': idx,
                    'ETag': part_id,
                }
                for idx, part_id in enumerate(self.part_ids, start=1)
            ]
        }

        # Combine the parts into a single S3 object
        self.s3.complete_multipart_upload(
            Bucket=self.bucket,
            Key=self.filename,
            UploadId=self.multipart_id,
            MultipartUpload=part_info
        )

        # Clear now invalid IDs
        self.part_ids = []
        self.multipart_id = None

    def close(self):
        # Ensure any locally buffered data is uploaded to a part
        while len(self.buffer):
            self._flush_buffer()

        # If the multipart upload was initialized, finalize it.
        if self.multipart_id is not None:
            self._finalize_multipart()

    def abort(self):
        """
        Use if for some reason you want to discard all the data written and not create an S3 object
        """
        self.buffer.clear()
        if self.multipart_id is None:
            # If the multipart upload was never initialized, do nothing
            return
        self.s3.abort_multipart_upload(
            Bucket=self.bucket,
            Key=self.filename,
            UploadId=self.multipart_id,
        )
        self.part_ids = []
        self.multipart_id = None

    def write(self, data: bytes):
        self.buffer.extend(data)
        while len(self.buffer) >= self.chunk_size:
            # _flush_buffer uploads only one part so multiple calls may be needed for large writes
            self._flush_buffer()


class S3Storage(StorageBackend):
    def __init__(
            self,
            bucket,
            aws_region,
            aws_access_key_id=None,
            aws_secret_access_key=None,
            aws_profile=None,
            name='s3'
    ):
        super().__init__()
        self.name = name
        self.bucket = bucket
        self.session = boto3.session.Session(
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            profile_name=aws_profile,
            region_name=aws_region
        )
        self.client = self.session.client('s3')

    def list(self, path):
        results = []

        more_results = True
        params = {
            'Bucket': self.bucket,
            'Prefix': path,
        }

        # The list_objects_v2 endpoint may paginate the results if there are a bunch of objects
        # that match the prefix
        while more_results:
            resp = self.client.list_objects_v2(**params)
            results.extend([
                ListEntry(
                    name=obj['Key'],
                    last_modified=arrow.get(obj['LastModified']),
                    size=obj['Size']
                )
                for obj in resp.get('Contents', [])
            ])

            # When IsTruncated is true it indicates that we need to call the endpoint again with
            # the given ContinuationToken to get the next batch of object descriptions.
            # The continuation token is an opaque string that points to the next page of results.
            more_results = resp['IsTruncated']
            if more_results:
                params['ContinuationToken'] = resp['NextContinuationToken']

        return results

    def _create_reader(self, path):
        try:
            return S3Reader(self.bucket, path, self.client)
        except ClientError as err:
            if err.response['Error']['Code'] == 'NoSuchKey':
                raise FileNotFoundInStorageError(storage_type=self, filename=path)
            raise

    def _create_writer(self, path):
        return S3Writer(self.bucket, path, self.client)

    def open(self, path: str, mode: typing.Union[FileMode, str]):
        mode = FileMode.as_mode(mode)

        if mode & FileMode.read and mode & FileMode.write:
            raise NotImplementedError('Read+write mode not supported by the S3 backend')
        elif mode & FileMode.read:
            return self._create_reader(path)
        elif mode & FileMode.write:
            return self._create_writer(path)
        else:
            raise ValueError('Unsupported mode. Accepted modes are FileMode.read or FileMode.write')

    def delete(self, path):
        self.client.delete_object(
            Bucket=self.bucket,
            Key=path
        )



import base64
import datetime
import string
from io import BytesIO

import arrow
import mock
import pytest
from azure.storage.blob import (
    Blob,
    BlobProperties,
    BlobPrefix,
)

from keg_storage import (
    backends,
    base,
)


@mock.patch.object(backends.AzureStorage, '_create_service')
class TestAzureStorage:
    def create_storage(self, **kwargs):
        return backends.AzureStorage(**{
            'account': 'foo',
            'key': 'bar',
            'bucket': 'test',
            **kwargs,
        })

    def test_list(self, m_service):
        m_list = m_service.return_value.list_blobs

        def blob(name, last_modified, size):
            props = BlobProperties()
            props.last_modified = last_modified
            props.content_length = size
            return Blob(name=name, props=props)

        prefix = BlobPrefix()
        prefix.name = 'abc/'

        m_list.return_value = [
            prefix,
            blob('abc/foo.txt', datetime.datetime(2019, 10, 5, 1, 1, 1), 123),
            blob('abc/bar.txt', datetime.datetime(2019, 10, 5, 1, 1, 2), 321),
            blob('baz.txt', datetime.datetime(2019, 10, 5, 1, 1, 3), 100),
        ]
        storage = self.create_storage()
        results = storage.list('xyz')

        m_list.assert_called_once_with(
            container_name='test',
            prefix='xyz/',
            delimiter='/',
        )
        assert results == [
            base.ListEntry(name='abc/', last_modified=None, size=0),
            base.ListEntry(name='abc/foo.txt', last_modified=arrow.get(2019, 10, 5, 1, 1, 1), size=123),  # noqa
            base.ListEntry(name='abc/bar.txt', last_modified=arrow.get(2019, 10, 5, 1, 1, 2), size=321),  # noqa
            base.ListEntry(name='baz.txt', last_modified=arrow.get(2019, 10, 5, 1, 1, 3), size=100),
        ]

    def test_open_read_write(self, m_service):
        storage = self.create_storage()
        with pytest.raises(NotImplementedError) as exc:
            storage.open('foo', base.FileMode.read | base.FileMode.write)
        assert str(exc.value) == 'Read+write mode not supported by the Azure backend'

    def test_open_bad_mode(self, m_service):
        storage = self.create_storage()
        with pytest.raises(ValueError) as exc:
            storage.open('foo', base.FileMode(0))
        assert (
            str(exc.value) == 'Unsupported mode. Accepted modes are FileMode.read or FileMode.write'
        )

    def test_read_operations(self, m_service):
        storage = self.create_storage()

        data = ''.join([
            string.digits,
            string.ascii_lowercase,
            string.ascii_uppercase,
        ]).encode()
        chunks = [data[i: i+10] for i in range(0, len(data), 10)]

        props = BlobProperties()
        props.content_length = len(data)
        m_service.return_value.get_blob_properties.return_value = Blob(props=props)

        m_get = m_service.return_value.get_blob_to_bytes
        m_get.side_effect = [Blob(content=c) for c in chunks]

        with storage.open('foo', base.FileMode.read, buffer_size=10) as f:
            assert f.blob_size == 62
            assert f.read(1) == b'0'
            m_get.assert_called_with(
                container_name='test',
                blob_name='foo',
                start_range=0,
                end_range=9,
            )

            assert f.read(2) == b'12'
            assert f.read(10) == b'3456789abc'
            m_get.assert_called_with(
                container_name='test',
                blob_name='foo',
                start_range=10,
                end_range=19,
            )
            assert f.read(30) == b'defghijklmnopqrstuvwxyzABCDEFG'
            m_get.assert_any_call(
                container_name='test',
                blob_name='foo',
                start_range=20,
                end_range=29,
            )
            m_get.assert_any_call(
                container_name='test',
                blob_name='foo',
                start_range=30,
                end_range=39,
            )
            m_get.assert_any_call(
                container_name='test',
                blob_name='foo',
                start_range=40,
                end_range=49,
            )

            assert f.read(30) == b'HIJKLMNOPQRSTUVWXYZ'
            m_get.assert_any_call(
                container_name='test',
                blob_name='foo',
                start_range=50,
                end_range=59,
            )
            m_get.assert_any_call(
                container_name='test',
                blob_name='foo',
                start_range=60,
                end_range=69,
            )

    @mock.patch('keg_storage.backends.azure.os.urandom', autospec=True, spec_set=True)
    def test_write_operations(self, m_urandom, m_service):
        storage = self.create_storage()

        m_urandom.side_effect = lambda x: b'\x00' * x

        block_data = {}
        blob = BytesIO()

        def mock_put_block(**kwargs):
            block_id = kwargs['block_id']
            assert block_id not in block_data
            block_data[block_id] = kwargs['block']

        def mock_put_list(**kwargs):
            blocks = kwargs['block_list']
            for b in blocks:
                blob.write(block_data[b.id])

        m_put_block = m_service.return_value.put_block
        m_put_block.side_effect = mock_put_block

        m_put_list = m_service.return_value.put_block_list
        m_put_list.side_effect = mock_put_list

        def block_id(index_bytes):
            return base64.b64encode(index_bytes + bytes([0] * 40)).decode()

        with storage.open('foo', base.FileMode.write, buffer_size=10) as f:
            f.write(b'ab')
            m_put_block.assert_not_called()
            m_put_list.assert_not_called()
            assert blob.getvalue() == b''
            assert block_data == {}

            f.write(b'cdefghijklm')
            m_put_block.assert_called_once_with(
                container_name='test',
                blob_name='foo',
                block=b'abcdefghij',
                block_id=block_id(b'\x00\x00\x00\x00\x00\x00\x00\x00')
            )
            m_put_list.assert_not_called()
            assert blob.getvalue() == b''
            assert set(block_data.values()) == {b'abcdefghij'}

            f.write(b'nopqrstuvwxyz')
            m_put_block.assert_called_with(
                container_name='test',
                blob_name='foo',
                block=b'klmnopqrst',
                block_id=block_id(b'\x00\x00\x00\x00\x00\x00\x00\x01')
            )
            m_put_block.reset_mock()
            m_put_list.assert_not_called()
            assert blob.getvalue() == b''
            assert set(block_data.values()) == {b'abcdefghij', b'klmnopqrst'}

            f.write(b'12')
            m_put_block.assert_not_called()
            m_put_list.assert_not_called()
            assert blob.getvalue() == b''
            assert set(block_data.values()) == {b'abcdefghij', b'klmnopqrst'}

        m_put_block.assert_called_with(
            container_name='test',
            blob_name='foo',
            block=b'uvwxyz12',
            block_id=block_id(b'\x00\x00\x00\x00\x00\x00\x00\x02')
        )
        m_put_list.assert_called_once()
        _, kwargs = m_put_list.call_args
        assert kwargs['container_name'] == 'test'
        assert kwargs['blob_name'] == 'foo'
        assert [b.id for b in kwargs['block_list']] == [
            block_id(b'\x00\x00\x00\x00\x00\x00\x00\x00'),
            block_id(b'\x00\x00\x00\x00\x00\x00\x00\x01'),
            block_id(b'\x00\x00\x00\x00\x00\x00\x00\x02')
        ]
        assert blob.getvalue() == b'abcdefghijklmnopqrstuvwxyz12'

    def test_write_nothing(self, m_service):
        storage = self.create_storage()
        m_put_block = m_service.return_value.put_block
        m_put_list = m_service.return_value.put_block_list

        f = storage.open('foo', base.FileMode.write)
        f.close()

        m_put_block.assert_not_called()
        m_put_list.assert_not_called()

    def test_delete(self, m_service):
        storage = self.create_storage()
        storage.delete('foo')

        m_service.return_value.delete_blob.assert_called_once_with(
            container_name='test',
            blob_name='foo'
        )



import click.testing
from blazeutils.containers import LazyDict
from datetime import date
from flask import current_app
from keg.testing import CLIBase, invoke_command
from mock import mock

from keg_storage import FileNotFoundInStorageError


@mock.patch.object(current_app.storage, 'get_interface', autospec=True, spec_set=True)
class TestCLI(CLIBase):
    def test_no_location(self, m_get_interface):
        results = self.invoke('storage', 'get', 'foo/bar', exit_code=1)
        assert results.output == 'No location given and no default was configured.\nAborted!\n'

        self.invoke('storage', '--location', 'loc', 'get', 'foo/bar')
        m_get_interface.assert_called_once_with('loc')

    def test_bad_location(self, m_get_interface):
        m_get_interface.side_effect = KeyError
        results = self.invoke('storage', '--location', 'loc', 'get', 'foo/bar', exit_code=1)
        assert results.output == 'The location loc does not exist. ' \
            'Pass --location or change your configuration.\nAborted!\n'


@mock.patch.object(current_app.storage, 'get_interface', autospec=True, spec_set=True)
class TestList(CLIBase):
    cmd_name = 'storage --location loc list'

    def test_list(self, m_get_interface):
        m_list = mock.MagicMock(
            return_value=[LazyDict(last_modified=date(2017, 4, 1), name='foo/bar', size=1024 * 3)]
        )
        m_get_interface.return_value.list = m_list

        results = self.invoke('foo')
        assert results.output == 'Apr 01 2017\t3.0K\tfoo/bar\n'
        m_list.assert_called_once_with('foo')

    def test_list_simple(self, m_get_interface):
        m_list = mock.MagicMock(
            return_value=[LazyDict(last_modified=date(2017, 4, 1), name='foo/bar', size=1024 * 3)]
        )
        m_get_interface.return_value.list = m_list

        results = self.invoke('--simple')
        assert results.output == 'foo/bar\n'
        m_list.assert_called_once_with('/')


@mock.patch.object(current_app.storage, 'get_interface', autospec=True, spec_set=True)
class TestGet(CLIBase):
    cmd_name = 'storage --location loc get'

    def test_get_default_dest(self, m_get_interface):
        m_get = mock.MagicMock()
        m_get_interface.return_value.get = m_get

        results = self.invoke('foo/bar')
        assert results.output == 'Downloaded foo/bar to bar.\n'
        m_get.assert_called_once_with('foo/bar', 'bar')

    def test_get_given_dest(self, m_get_interface):
        m_get = mock.MagicMock()
        m_get_interface.return_value.get = m_get

        results = self.invoke('foo/bar', 'dest/path')
        assert results.output == 'Downloaded foo/bar to dest/path.\n'
        m_get.assert_called_once_with('foo/bar', 'dest/path')

    def test_get_file_not_found(self, m_get_interface):
        m_get_interface.return_value.get.side_effect = FileNotFoundInStorageError('abc', 'def')

        results = self.invoke('foo/bar', exit_code=1)
        assert results.output == 'Error: Could not open file def: Not found in abc.\n'


@mock.patch.object(current_app.storage, 'get_interface', autospec=True, spec_set=True)
class TestPut(CLIBase):
    cmd_name = 'storage --location loc put'

    def test_put(self, m_get_interface):
        m_put = mock.MagicMock()
        m_get_interface.return_value.put = m_put

        results = self.invoke('foo/bar', 'baz/bar')
        assert results.output == 'Uploaded foo/bar to baz/bar.\n'
        m_put.assert_called_once_with('foo/bar', 'baz/bar')


@mock.patch.object(current_app.storage, 'get_interface', autospec=True, spec_set=True)
class TestDelete(CLIBase):
    cmd_name = 'storage --location loc delete'

    def test_delete(self, m_get_interface):
        m_delete = mock.MagicMock()
        m_get_interface.return_value.delete = m_delete

        results = self.invoke('foo/bar')
        assert results.output == 'Deleted foo/bar.\n'
        m_delete.assert_called_once_with('foo/bar')

    def test_delete_file_not_found(self, m_get_interface):
        m_get_interface.return_value.delete.side_effect = FileNotFoundInStorageError('abc', 'def')

        results = self.invoke('foo/bar', exit_code=1)
        assert results.output == 'Error: Could not open file def: Not found in abc.\n'


@mock.patch.object(current_app.storage, 'get_interface', autospec=True, spec_set=True)
class TestLinkFor(CLIBase):
    cmd_name = 'storage --location loc link_for'

    def test_link_for_default_expiration(self, m_get_interface):
        m_link = mock.MagicMock(return_value='http://example.com/foo/bar')
        m_get_interface.return_value.link_for = m_link

        results = self.invoke('foo/bar')
        assert results.output == 'http://example.com/foo/bar\n'
        m_link.assert_called_once_with('foo/bar', 3600)

    def test_link_for_expiration_given(self, m_get_interface):
        m_link = mock.MagicMock(return_value='http://example.com/foo/bar')
        m_get_interface.return_value.link_for = m_link

        results = self.invoke('-e', '100', 'foo/bar')
        assert results.output == 'http://example.com/foo/bar\n'
        m_link.assert_called_once_with('foo/bar', 100)

    def test_link_for_error(self, m_get_interface):
        m_get_interface.return_value.link_for.side_effect = ValueError('Some error')

        results = self.invoke('foo/bar', exit_code=1)
        assert results.output == 'Some error\nAborted!\n'


@mock.patch.object(current_app.storage, 'get_interface', autospec=True, spec_set=True)
@mock.patch('keg_storage.utils.reencrypt', autospec=True, spec_set=True)
class TestReencrypt(CLIBase):
    class InputRunner(click.testing.CliRunner):
        def isolation(self, **kwargs):
            kwargs['input'] = 'abc\nxyz'
            return super(self.__class__, self).isolation(**kwargs)

    def test_reencrypt(self, m_reencrypt, m_get_interface):
        m_get_interface.return_value = 'STORAGE'
        app_cls = current_app._get_current_object().__class__
        results = invoke_command(app_cls, 'storage', '--location', 'loc', 'reencrypt', 'foo',
                                 runner=self.InputRunner())

        assert results.output == 'Old Key: \nNew Key: \nRe-encrypted foo\n'
        m_reencrypt.assert_called_once_with('STORAGE', 'foo', b'abc', b'xyz')

    def test_file_not_found(self, m_reencrypt, m_get_interface):
        m_get_interface.return_value = 'STORAGE'
        m_reencrypt.side_effect = FileNotFoundInStorageError('abc', 'def')

        app_cls = current_app._get_current_object().__class__
        results = invoke_command(app_cls, 'storage', '--location', 'loc', 'reencrypt', 'foo',
                                 runner=self.InputRunner(), exit_code=1)

        assert results.output.splitlines() == [
            'Old Key: ',
            'New Key: ',
            'Error: Could not open file def: Not found in abc.'
        ]



import logging
import tempfile

from blazeutils.helpers import ensure_list

try:
    import keg_elements.crypto as ke_crypto
except ImportError:
    ke_crypto = None

DEFAULT_KEY_SIZE = 32

log = logging.getLogger(__name__)


class DecryptionException(Exception):
    pass


class EncryptionKeyException(Exception):
    pass


class MissingDependencyException(Exception):
    pass


def verify_key_length(key, expected=DEFAULT_KEY_SIZE):
    return len(key) == expected


def reencrypt(storage, path, old_key, new_key):
    if ke_crypto is None:
        raise MissingDependencyException('Keg Elements is required for crypto operations')

    old_key = ensure_list(old_key)

    # We append the new key just in case the operation was restarted and we have already encrypted
    # some files
    keys = list(filter(verify_key_length, old_key + [new_key]))

    if not keys:
        raise EncryptionKeyException('No Keys Found')
    else:
        log.info('Found {} keys'.format(len(keys)))

    if not verify_key_length(new_key, expected=DEFAULT_KEY_SIZE):
        raise EncryptionKeyException('New key is not the correct size. Got {}, expecting {}'.format(
            len(new_key), DEFAULT_KEY_SIZE
        ))

    with tempfile.NamedTemporaryFile() as old_key_local, \
            tempfile.NamedTemporaryFile() as new_key_local:

        log.info('Fetching {}'.format(path))
        storage.get(path, old_key_local.name)

        old_key_bytes = None
        for idx, key in enumerate(keys):
            try:
                log.info('Trying to decrypt {}.'.format(path, len(keys)))
                old_key_bytes = ke_crypto.decrypt_bytesio(key, old_key_local.name)
                log.info('Successfully Decrypted {} with key {}.'.format(path, idx))
            except Exception as e:
                log.info('Key {} failed for {}'.format(idx, path))

                if str(e) == 'Invalid padding bytes.':
                    continue

                log.error('Unhandled error for decrypt with key {}: {}.'.format(idx, str(e)))

        if old_key_bytes is None:
            raise DecryptionException('Unable to Decrypt File {}'.format(path))

        log.info('Re-encrypting {}'.format(path))
        new_key_bytes = ke_crypto.encrypt_fileobj(new_key, old_key_bytes)

        log.info('Writing newly encrypted data {}.'.format(path))
        for chunk in new_key_bytes:
            new_key_local.write(chunk)
        new_key_local.flush()

        storage.put(new_key_local.name, path)
        log.info('Re-encryption complete for {}.'.format(path))



from importlib.util import find_spec

from .base import StorageBackend, FileNotFoundInStorageError, FileMode  # noqa

__all__ = [
    'StorageBackend',
    'FileNotFoundInStorageError',
    'FileMode',
    'base',
]

if find_spec('boto3') is not None:
    from .s3 import S3Storage  # noqa
    __all__.extend(['s3', 'S3Storage'])

if find_spec('paramiko') is not None:
    from .sftp import SFTPStorage  # noqa
    __all__.extend(['sftp', 'SFTPStorage'])

if find_spec('azure') is not None:
    from .azure import AzureStorage  # noqa
    __all__.extend(['azure', 'AzureStorage'])



import os.path as osp
from setuptools import setup, find_packages


cdir = osp.abspath(osp.dirname(__file__))
README = open(osp.join(cdir, 'README.rst')).read()
CHANGELOG = open(osp.join(cdir, 'changelog.rst')).read()

version = {}
with open(osp.join(cdir, 'keg_storage', 'version.py')) as version_fp:
    exec(version_fp.read(), version)

setup(
    name="KegStorage",
    description="A simple storage interface with multiple backends for use in a Keg_ app.",
    long_description='\n\n'.join((README, CHANGELOG)),
    author="Level 12 Developers",
    author_email="devteam@level12.io",
    url='https://github.com/level12/keg-storage',
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    license='BSD',
    packages=find_packages(),
    zip_safe=True,
    version=version['VERSION'],
    install_requires=[
        'arrow',
        'boto3',
        'botocore',
        'humanize',
        'BlazeUtils',
    ],
    extras_require={
        'sftp': [
            'paramiko',
        ],
        'azure': [
            'azure',
        ],
        'keg': [
            'kegelements',
        ],
        'test': [
            'azure',
            'paramiko',
            'tox',
            'pytest',
            'pytest-coverage',
            'wrapt',
            'mock',
            'flake8',
            'flask_webtest',
        ],
    }
)



from mock import mock

import keg_storage


class TestStorage:
    def test_app_to_init_call_init(self):
        app = mock.MagicMock()
        app.config = {'STORAGE_PROFILES': [(keg_storage.backends.StorageBackend, {'name': 'test'})]}
        storage = keg_storage.Storage(app)
        assert 'test' in storage._interfaces
        assert storage.interface == 'test'



import base64
import os
import typing

import arrow
from azure.storage.blob import (
    BlockBlobService,
    BlobPrefix,
    BlobBlock,
)

from keg_storage.backends import base


class AzureFile(base.RemoteFile):
    """
    Base class for Azure file interface. Since read and write operations are very different and
    integrating the two would introduce a lot of complexity there are distinct subclasses for files
    opened for reading and writing.
    """
    def __init__(self, container_name: str, path: str, mode: base.FileMode,
                 service: BlockBlobService, chunk_size: int = 10 * 1024 * 1024):
        """
        :param container_name: container / bucket name
        :param path: blob name
        :param mode: file mode
        :param service: blob service instance to use for API calls
        :param chunk_size: read/write buffer size
        """
        super().__init__(mode)
        self.chunk_size = chunk_size
        self.container_name = container_name
        self.path = path
        self.service = service

        # Local buffer to reduce number of requests
        self.buffer = bytearray()


class AzureWriter(AzureFile):
    """
    We are using Azure Block Blobs for all operations. The process for writing them is substantially
    similar to that of S3 with a couple of differences.
        1. We generate the IDs for the blocks
        2. There is no separate call to instantiate the upload. The first call to put_block will
           create the blob.
    """
    def __init__(self, container_name: str, path: str, mode: base.FileMode,
                 service: BlockBlobService, chunk_size: int = BlockBlobService.MAX_BLOCK_SIZE):
        super().__init__(
            container_name=container_name,
            path=path,
            mode=mode,
            service=service,
            # chunk_size cannot be larger than MAX_BLOCK_SIZE due to API restrictions
            chunk_size=min(chunk_size, BlockBlobService.MAX_BLOCK_SIZE),
        )
        self.blocks = []

    def _gen_block_id(self) -> str:
        """
        Generate a unique ID for the block. This is meant to be opaque but it is generated from:
            1. The index of the block as an 64 bit unsigned big endian integer
            2. 40 bytes of random data
        The two parts are concatenated and base64 encoded giving us 64 bytes which is the maximum
        Azure allows.
        """
        index_part = len(self.blocks).to_bytes(8, byteorder='big', signed=False)
        random_part = os.urandom(40)
        return base64.b64encode(index_part + random_part).decode()

    def _flush(self):
        if len(self.buffer) == 0:
            # If there is no buffered data, we don't need to do anything
            return

        # Upload at most chunk_size bytes to a new block
        block_id = self._gen_block_id()
        self.service.put_block(
            container_name=self.container_name,
            blob_name=self.path,
            block=bytes(self.buffer[:self.chunk_size]),
            block_id=block_id
        )

        # Store the block_id to later concatenate when we close this file
        self.blocks.append(BlobBlock(id=block_id))

        # Cycle the buffer
        self.buffer = self.buffer[self.chunk_size:]

    def _finalize(self):
        self.service.put_block_list(
            container_name=self.container_name,
            blob_name=self.path,
            block_list=self.blocks
        )
        self.blocks = []

    def write(self, data: bytes) -> None:
        self.buffer.extend(data)
        while len(self.buffer) >= self.chunk_size:
            # Write may be bigger than the chunk size and _flush() only uploads a single chunk so
            # repeated calls may be necessary
            self._flush()

    def close(self):
        self._flush()
        if self.blocks:
            # If we haven't created any blocks, we don't need to finalize
            self._finalize()


class AzureReader(AzureFile):
    """
    The Azure reader uses byte ranged API calls to fill a local buffer to avoid lots of API overhead
    for small read sizes.
    """
    def __init__(self, container_name: str, path: str, mode: base.FileMode,
                 service: BlockBlobService, chunk_size=10 * 1024 * 1024):
        super().__init__(
            container_name=container_name,
            path=path,
            mode=mode,
            service=service,
            chunk_size=chunk_size,
        )
        # Get overall blob size to avoid out of bounds requests
        self.blob_size = self._get_blob_size()

        # Current offset for the next request to fill the local buffer
        self.current_offset = 0

        # Current read position in the blob including reads filled from the local buffer
        self.read_position = 0

    def _get_blob_size(self):
        """
        Retrieve the overall size of the blob
        """
        blob = self.service.get_blob_properties(
            container_name=self.container_name,
            blob_name=self.path
        )
        return blob.properties.content_length

    def _refill_buffer(self):
        """
        Retrieve the next chunk and add it to the local buffer.
        """
        range_end = self.current_offset + self.chunk_size - 1  # -1 because range is inclusive
        blob = self.service.get_blob_to_bytes(
            container_name=self.container_name,
            blob_name=self.path,
            start_range=self.current_offset,
            end_range=range_end,
        )
        self.current_offset += len(blob.content)
        self.buffer.extend(blob.content)

    def _read_from_buffer(self, max_size):
        """
        Read up to max_size bytes from the local buffer.
        """
        read_size = min(len(self.buffer), max_size)
        output = self.buffer[:read_size]
        self.buffer = self.buffer[read_size:]
        self.read_position += len(output)
        return output

    def read(self, size: int) -> bytes:
        output_buf = bytes()

        # Read up to the end of the file or `size` bytes whichever is less
        read_size = min(self.blob_size - self.read_position, size)

        while len(output_buf) < read_size:
            if len(self.buffer) == 0:
                # Buffer is empty, refill it.
                self._refill_buffer()

            remaining = read_size - len(output_buf)
            read = self._read_from_buffer(remaining)
            output_buf += read

        return output_buf


class AzureStorage(base.StorageBackend):
    def __init__(self, account: str, key: str, bucket: str, name: str = 'azure'):
        super().__init__()
        self.name = name
        self.account = account
        self.key = key
        self.bucket = bucket

    def _create_service(self):
        return BlockBlobService(
            account_name=self.account,
            account_key=self.key,
        )

    def list(self, path: str) -> typing.List[base.ListEntry]:
        service = self._create_service()
        if not path.endswith('/'):
            path = path + '/'
        list_iter = service.list_blobs(container_name=self.bucket, prefix=path, delimiter='/')

        def construct_entry(blob):
            if isinstance(blob, BlobPrefix):
                return base.ListEntry(name=blob.name, last_modified=None, size=0)
            return base.ListEntry(
                name=blob.name,
                last_modified=arrow.get(blob.properties.last_modified),
                size=blob.properties.content_length,
            )

        return [construct_entry(blob) for blob in list_iter]

    def open(self, path: str, mode: typing.Union[base.FileMode, str],
             buffer_size: int = 10 * 1024 * 1024) -> AzureFile:  # noqa
        mode = base.FileMode.as_mode(mode)

        if (mode & base.FileMode.read) and (mode & base.FileMode.write):
            raise NotImplementedError('Read+write mode not supported by the Azure backend')
        elif mode & base.FileMode.write:
            return AzureWriter(
                container_name=self.bucket,
                path=path,
                mode=mode,
                service=self._create_service(),
                chunk_size=buffer_size
            )
        elif mode & base.FileMode.read:
            return AzureReader(
                container_name=self.bucket,
                path=path,
                mode=mode,
                service=self._create_service(),
                chunk_size=buffer_size
            )
        else:
            raise ValueError('Unsupported mode. Accepted modes are FileMode.read or FileMode.write')

    def delete(self, path: str):
        service = self._create_service()
        service.delete_blob(
            container_name=self.bucket,
            blob_name=path
        )



import enum
import typing

import arrow


class ListEntry(typing.NamedTuple):
    name: str
    last_modified: arrow.Arrow
    size: int


class FileMode(enum.Flag):
    read = enum.auto()
    write = enum.auto()

    def __str__(self):
        s = 'r' if self & FileMode.read else ''
        s += 'w' if self & FileMode.write else ''
        return f'{s}b'

    @classmethod
    def as_mode(cls, obj):
        if isinstance(obj, cls):
            return obj
        if not isinstance(obj, str):
            raise ValueError('as_mode() accepts only FileMode or str arguments')

        mode = cls(0)
        if 'r' in obj:
            mode |= cls.read
        if 'w' in obj:
            mode |= cls.write
        return mode


class RemoteFile:
    """
    This is a base class for objects returned by a backend's `open()` method. This is a file-like
    object that provides read/write operations to the remote file. When creating a new backend, you
    should subclass this and implement `read()` and `write()` methods at minimum.

    After construction, a RemoteFile is presumed to be in an "open" state and should accept calls
    to any of its methods. Any cleanup should be done in the `close()` method.
    """

    # This is the default chunk size to use when iterating over this object
    iter_chunk_size = 5 * 1024 * 1024

    def __init__(self, mode: FileMode):
        """
        Override this constructor to accept any additional arguments needed by the backend and to
        perform any initialization required to get the file into an "open" state.
        """
        self.mode = mode

    def read(self, size: int) -> bytes:
        """
        Read and return up to `size` bytes from the remote file. If the end of the file is reached
        this should return an empty bytes string.
        """
        raise NotImplementedError

    def write(self, data: bytes) -> None:
        """
        Write the data buffer to the remote file.
        """
        raise NotImplementedError

    def close(self):
        """
        Cleanup and deallocate any held resources. This method may be called multiple times on a
        single instance. If the file was already closed, this method should do nothing.
        """
        pass

    def iter_chunks(self, chunk_size: int = None):
        """
        Iterate over the file in blocks of `chunk_size`.
        """
        chunk_size = chunk_size or self.iter_chunk_size

        while True:
            chunk = self.read(chunk_size)
            if chunk == b'':
                break
            yield chunk

    def __iter__(self):
        return self.iter_chunks()

    def __del__(self):
        self.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class StorageBackend:
    name = None

    def __init__(self, *args, **kwargs):
        pass

    def list(self, path: str) -> typing.List[ListEntry]:
        """
        Returns a list of `ListEntry`s representing files available under the directory or prefix
        given in `path`.
        """
        raise NotImplementedError()

    def open(self, path: str, mode: typing.Union[FileMode, str]) -> RemoteFile:
        """
        Returns a instance of RemoteFile for the given `path` that can be used for
        reading and/or writing depending on the `mode` given.
        """
        raise NotImplementedError()

    def delete(self, path: str):
        """
        Delete the remote file specified by `path`.
        """
        raise NotImplementedError()

    def get(self, path: str, dest: str) -> None:
        """
        Copies a remote file at `path` to the `dest` path given on the local filesystem.
        """
        with self.open(path, FileMode.read) as infile, open(dest, str(FileMode.write)) as outfile:
            for chunk in infile.iter_chunks():
                outfile.write(chunk)

    def put(self, path: str, dest: str) -> None:
        """
        Copies a local file at `path` to a remote file at `dest`.
        """
        buffer_size = 5 * 1024 * 1024

        with self.open(dest, FileMode.write) as outfile, open(path, str(FileMode.read)) as infile:
            buf = infile.read(buffer_size)
            while buf:
                outfile.write(buf)
                buf = infile.read(buffer_size)

    def __str__(self):
        return self.__class__.__name__


class FileNotFoundInStorageError(Exception):
    def __init__(self, storage_type, filename):
        self.storage_type = storage_type
        self.filename = filename

    def __str__(self):
        return "File {} not found in {}.".format(self.filename, str(self.storage_type))
